/*
 * Copyright (c) 2004 Sean Chittenden <sean@chittenden.org>
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * $Nexadesic: src/lib/libmemcache/memcache.c,v 1.4 2004/12/08 00:25:19 sean Exp $
 */

/* The crc32 functions and data was originally written by Spencer
 * Garrett <srg@quick.com> and was cleaned from the PostgreSQL source
 * tree via the files contrib/ltree/crc32.[ch].  No license was
 * included, therefore it is assumed that this code is public
 * domain. */

#include <ctype.h>
#include <err.h>
#include <string.h>
#include <strings.h>
#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <unistd.h>
#include <fcntl.h>

#include "memcache.h"

static void			mc_fetch_cmd(struct memcache *mc, struct memcache_req *req,
					     const char *cmd, const size_t cmd_len);
static char			*mc_get_line(struct memcache *mc, struct memcache_server *ms);
static void			 mc_server_block(struct memcache_server *ms, int block);
static void			 mc_reset_buf(struct memcache *mc);
static struct memcache_res	*mc_res_new(void);
static int			 mc_server_connect(struct memcache *mc, struct memcache_server *ms);
static void			 mc_server_free(struct memcache_server *ms);
static struct memcache_server	*mc_server_new(struct memcache *mc);
static int			 mc_server_resolve(struct memcache_server *ms);
static struct memcache_server_stats	*mc_server_stats_new(void);
static int			 mc_storage_cmd(struct memcache *mc,
						const char *cmd, const size_t cmd_len,
						const char *key, const size_t key_len,
						const void *val, const size_t bytes,
						const time_t expire, const u_int16_t flags);
static u_int32_t		 mc_atomic_cmd(struct memcache *mc, const char *cmd, const size_t cmd_len,
					       const char *key, const size_t key_len, const u_int32_t val);


#ifdef __STRICT_ANSI__
long long strtoll(const char *, char **, int);
#endif

/* This is kinda ugly, but, it saves on some warnings and a tad of
 * stack space across the library. Note, remember strlen(3) does not
 * include the trailing null character, but sizeoof() does, so when
 * computing the sizeof() commands, subtract one from its return. */
const char	str_add_cmd[] = "add ";
const size_t	str_add_len = sizeof(str_add_cmd) - 1;
const char	str_decr_cmd[] = "decr ";
const size_t	str_decr_len = sizeof(str_decr_cmd) - 1;
const char	str_delete_cmd[] = "delete ";
const size_t	str_delete_len = sizeof(str_delete_cmd) - 1;
const char	str_endl[] = "\r\n";
const size_t	str_endl_len = sizeof(str_endl) - 1;
const char	str_get_cmd[] = "get ";
const size_t	str_get_len = sizeof(str_get_cmd) - 1;
const char	str_incr_cmd[] = "incr ";
const size_t	str_incr_len = sizeof(str_incr_cmd) - 1;
const char	str_refresh_cmd[] = "refresh ";
const size_t	str_refresh_len = sizeof(str_refresh_cmd) - 1;
const char	str_replace_cmd[] = "replace ";
const size_t	str_replace_len = sizeof(str_replace_cmd) - 1;
const char	str_set_cmd[] = "set ";
const size_t	str_set_len = sizeof(str_set_cmd) - 1;
const char	str_space[] = " ";
const size_t	str_space_len = sizeof(str_space) - 1;

/* #ifdef HAVE_KQUEUE */
/* /\* kqueue(2) is the preferred method for testing to see if a file */
/*  * descriptor has data on it. *\/ */
/* #else */
/* # ifdef HAVE_POLL */
/* /\* poll(2) is better than select(2), but still sucks*\/ */
/* # else */
/* #  ifdef HAVE_SELECT */
/* /\* If we have to, we'll use select(2). *\/ */
/* #  endif */
/* # endif */
/* #endif */

/* BEGIN memcache memory API */

/* Set the default memory handling routines to be system defaults. */
mcFreeFunc	mcFree		= (mcFreeFunc)free;
mcMallocFunc	mcMalloc	= (mcMallocFunc)malloc;
mcReallocFunc	mcRealloc	= (mcReallocFunc)realloc;
mcStrdupFunc	mcStrdup	= (mcStrdupFunc)strdup;

int
mcMemSetup(mcFreeFunc freeFunc, mcMallocFunc mallocFunc,
	   mcReallocFunc reallocFunc, mcStrdupFunc strdupFunc) {
  if (freeFunc == NULL || mallocFunc == NULL || reallocFunc == NULL || strdupFunc == NULL)
    return(1);
  mcFree = freeFunc;
  mcMalloc = mallocFunc;
  mcRealloc = reallocFunc;
  mcStrdup = strdupFunc;
  return(0);
}


int
mcMemGet(mcFreeFunc *freeFunc, mcMallocFunc *mallocFunc,
	 mcReallocFunc *reallocFunc, mcStrdupFunc *strdupFunc) {
  if (freeFunc != NULL)
    *freeFunc = mcFree;

  if (mallocFunc != NULL)
    *mallocFunc = mcMalloc;

  if (reallocFunc != NULL)
    *reallocFunc = mcRealloc;

  if (strdupFunc != NULL)
    *strdupFunc = mcStrdup;

  return(0);
}

/* END memcache memory API */


int
mc_add(struct memcache *mc,
       const char *key, const size_t key_len,
       const void *val, const size_t bytes,
       const time_t expire, const u_int16_t flags) {
  return mc_storage_cmd(mc, str_add_cmd, str_add_len, key, key_len, val, bytes, expire, flags);
}


/* Issues a "get" command to the memcache server that should contain
 * the key.  The result is mcMalloc(3)'ed and it is assumed that the
 * caller is required to mcFree(3) the memory. */
void *
mc_aget(struct memcache *mc, const char *key, const size_t len) {
  struct memcache_req *req;
  struct memcache_res *res;
  void *ret;

  req = mc_req_new();
  res = mc_req_add(req, key, len);
  mc_res_free_on_delete(res, 0);
  mc_get(mc, req);
  ret = res->val;
  mc_req_free(req);
  return ret;
}


/* Issues a "refresh" command to the memcache server that should
 * contain the key.  The result is mcMalloc(3)'ed and it is assumed
 * that the caller is required to mcFree(3) the memory. */
void *
mc_arefresh(struct memcache *mc, const char *key, const size_t len) {
  struct memcache_req *req;
  struct memcache_res *res;
  void *ret;

  req = mc_req_new();
  res = mc_req_add(req, key, len);
  mc_res_free_on_delete(res, 0);
  mc_refresh(mc, req);
  ret = res->val;
  mc_req_free(req);
  return ret;
}


static u_int32_t
mc_atomic_cmd(struct memcache *mc, const char *cmd, const size_t cmd_len,
	      const char *key, const size_t key_len, const u_int32_t val) {
  struct memcache_server *ms;
  u_int32_t hash;
  char *cp, *cur;
  size_t buf_left, i;
  struct iovec av[5];
  u_int32_t ret;

  hash = mc_hash_key(key, key_len);
  ms = mc_find_server(mc, hash);
  if (ms == NULL) {
    warnx("%s:%u\tUnable to find a valid server", __FILE__, __LINE__);
    return 0;
  }

  if (mc_server_connect(mc, ms) == -1)
    return 0;

  mc_reset_buf(mc);
  cur = mc->buf;
  buf_left = mc->size;

  av[0].iov_base = cmd;
  av[0].iov_len = cmd_len;
  av[1].iov_base = key;
  av[1].iov_len = key_len;
  av[2].iov_base = str_space;
  av[2].iov_len = str_space_len;

  /* Convert the value to a string */
  i = (size_t)snprintf(cur, buf_left, "%u", val);
  if (i < 1)
    err(EX_SOFTWARE, "%s:%u\tsnprintf()", __FILE__, __LINE__);

  av[3].iov_base = cur;
  av[3].iov_len = i;
  av[4].iov_base = str_endl;
  av[4].iov_len = str_endl_len;

  if (writev(ms->fd, av, 5) < 0) {
    warn("%s:%u\twritev()", __FILE__, __LINE__);
    mc_deactivate_server(mc, ms);
    /* XXX Should we recursively attempt to try this query on the
     * remaining servers in the cluster if the writev() fails?
     * Eventually we'd fail once all servers were exhausted?  For now,
     * just fail and return 0. */
    return 0;
  }

  mc_server_block(ms, 1);
  mc_reset_buf(mc);
  cur = mc_get_line(mc, ms);
  if (cur != NULL && memcmp(cur, "NOT_FOUND", sizeof("NOT_FOUND") - 1) == 0) {
    mc_server_block(ms, 0);
    return 0;
  } else if (cur == NULL) {
    return 0;
  }

  /* Try converting the value to an integer. If it succeeds, we've got
   * a winner. */
  ret = (u_int32_t)strtol(cur, &cp, 10);
  if (ret == 0 && (errno == EINVAL || errno == ERANGE))
    err(EX_PROTOCOL, "%s:%u\tstrtol(): invalid value \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);

#ifdef PEDANTIC
  if (*cp != '\0')
    errx(EX_PROTOCOL, "%s:%u\tProtocol error: %u", __FILE__, __LINE__, *cp);
#endif

  return ret;
}


void
mc_deactivate_server(struct memcache *mc, struct memcache_server *ms) {
  u_int32_t i, found;

  /* Since adding servers is so rare, and servers do come back, don't
   * bother mcRealloc(3)'ing mc->live_servers.  Instead, just find the
   * dead server in the array, remove it, shift the remaining servers
   * down a place in the array, and decrement number of servers in the
   * array.  This should making adding servers back to the list more
   * efficient. */
  for (i = 0, found = 0; i < mc->num_live_servers; i++) {
    if (found == 0) {
      if (mc->live_servers[i] == ms)
	found = 1;
      else
	continue;
    }

    mc->live_servers[i] = mc->live_servers[i + 1];
  }

  if (found != 0) {
    mc->num_live_servers--;
    /* If we've still got the server marked up, then set it to down.
     * If we've already marked it as down, keep the original reason
     * since it's going to be more verbose than saying 'down'.  A
     * 'down' server can be resurected, where as a server that's
     * marked 'no host' won't ever be resurected. */
    if (ms->active == 'u')
      ms->active = 'd';

    if (ms->fd != -1) {
      if (close(ms->fd) != 0)
	warn("%s:%u\tclose()", __FILE__, __LINE__);
      ms->fd = -1;
    }
  }
}


u_int32_t
mc_decr(struct memcache *mc, const char *key, const size_t key_len, const u_int32_t val) {
  return mc_atomic_cmd(mc, str_decr_cmd, str_decr_len, key, key_len, val);
}


int
mc_delete(struct memcache *mc, const char *key, const size_t key_len, const time_t hold) {
  struct memcache_server *ms;
  u_int32_t hash;
  char *cur;
  size_t buf_left, i;
  struct iovec dv[5];

  hash = mc_hash_key(key, key_len);
  ms = mc_find_server(mc, hash);
  if (ms == NULL) {
    warnx("%s:%u\tUnable to find a valid server", __FILE__, __LINE__);
    return -1;
  }

  if (mc_server_connect(mc, ms) == -1)
    return -2;

  mc_reset_buf(mc);
  cur = mc->buf;
  buf_left = mc->size;

  dv[0].iov_base = str_delete_cmd;
  dv[0].iov_len = str_delete_len;
  dv[1].iov_base = key;
  dv[1].iov_len = key_len;
  dv[2].iov_base = str_space;
  dv[2].iov_len = str_space_len;

  /* Convert expiration time to a string */
  i = (size_t)snprintf(cur, buf_left, "%lu", hold);
  if (i < 1)
    err(EX_SOFTWARE, "%s:%u\tsnprintf()", __FILE__, __LINE__);

  dv[3].iov_base = cur;		/* Note where our flags string is located */
  dv[3].iov_len = i;
  dv[4].iov_base = str_endl;
  dv[4].iov_len = str_endl_len;

  if (writev(ms->fd, dv, 5) < 0) {
    warn("%s:%u\twritev()", __FILE__, __LINE__);
    mc_deactivate_server(mc, ms);
    /* XXX Should we recursively attempt to try this query on the
     * remaining servers in the cluster if the writev() fails?
     * Eventually we'd fail once all servers were exhausted?  For now,
     * just fail and return NULL. */
    return -3;
  }

  mc_server_block(ms, 1);
  mc_reset_buf(mc);
  cur = mc_get_line(mc, ms);
  if (cur != NULL && memcmp(cur, "DELETED", sizeof("DELETED") - 1) == 0) {
    mc_server_block(ms, 0);
    return 0;
  } else if (cur != NULL && memcmp(cur, "NOT_FOUND", sizeof("NOT_FOUND") - 1) == 0) {
    mc_server_block(ms, 0);
    return 1;
  } else {
    errx(EX_PROTOCOL, "%s:%u\tProtocol error", __FILE__, __LINE__);
  }
}


struct memcache_server *
mc_find_server(struct memcache *mc, const u_int32_t hash) {
  if (mc->num_live_servers < 1)
    return NULL;

  /* Grab the correct server from the list. */
  return mc->live_servers[hash % mc->num_live_servers];
}


/* xxx This should really walk all servers */
int
mc_flush_all(struct memcache *mc, const char *key, const size_t key_len) {
  struct memcache_server *ms;
  u_int32_t hash;
  char *cur;

  hash = mc_hash_key(key, key_len);
  ms = mc_find_server(mc, hash);
  if (ms == NULL) {
    warnx("%s:%u\tUnable to find a valid server", __FILE__, __LINE__);
    return -1;
  }

  if (mc_server_connect(mc, ms) == -1)
    return -2;

  if (write(ms->fd, "flush_all\r\n", sizeof("flush_all\r\n") - 1) < 0) {
    warn("%s:%u\twrite()", __FILE__, __LINE__);
    mc_deactivate_server(mc, ms);
    /* XXX Should we recursively attempt to try this query on the
     * remaining servers in the cluster if the write() fails?
     * Eventually we'd fail once all servers were exhausted?  For now,
     * just fail and return NULL. */
    return -3;
  }

  mc_server_block(ms, 1);
  mc_reset_buf(mc);
  cur = mc_get_line(mc, ms);
  if (cur != NULL && memcmp(cur, "OK", sizeof("OK") - 1) == 0) {
    mc_server_block(ms, 0);
    return 0;
  } else {
    errx(EX_PROTOCOL, "%s:%u\tProtocol error", __FILE__, __LINE__);
  }
}


void
mc_free(struct memcache *mc) {
  struct memcache_server *ms, *tms;

  if (mc == NULL)
    return;

  for (ms = mc->server_list.tqh_first; ms != NULL; ms = ms->entries.tqe_next) {
    tms = ms->entries.tqe_next;

    mc_server_free(ms);

    if (tms != NULL)
      ms = tms;
  }

  if (mc->size > 0)
    mcFree(mc->buf);

  mcFree(mc);
}


static void
mc_fetch_cmd(struct memcache *mc, struct memcache_req *req, const char *cmd, const size_t cmd_len) {
  struct memcache_res *res;
  struct memcache_server *ms;
  struct iovec *rv;
  u_int32_t i, num_vec;
  u_int16_t flags;
  char *cp, *cur;
  size_t bytes;
  ssize_t rb;

  if (req->num_keys == 0)
    return;

  /* Need two extra iovec structs, one for the command, one for the
   * trailing \r\n. */
  num_vec = 2 * req->num_keys + 1;
  rv = (struct iovec *)mcMalloc(sizeof(struct iovec) * num_vec);
  rv[0].iov_base = cmd;
  rv[0].iov_len = cmd_len;
  for (i = 1, res = req->query.tqh_first; res != NULL; res = res->entries.tqe_next, i++) {
    if (res->hash == 0)
      res->hash = mc_hash_key(res->key, res->len);

    rv[i].iov_base = res->key;
    rv[i].iov_len = res->len;

    if (res->entries.tqe_next != NULL) {
      rv[++i].iov_base = str_space;
      rv[i].iov_len = 1;
    }

    /* While we're looping, might as well see if we should be auto
     * deleting any of these keys. */
    if (res->_flags & (MC_RES_FREE_ON_DELETE | MC_RES_NO_FREE_ON_DELETE) ==
	(MC_RES_FREE_ON_DELETE | MC_RES_FREE_ON_DELETE))
      mc_res_free_on_delete(res, (res->size > 0 ? 0 : 1));
  }
  rv[i].iov_base = str_endl;
  rv[i].iov_len = str_endl_len;

  /* Yuk.  Right now I'm grabbing the hash from the first key and
   * using that, but that's dangerous in the event that the second key
   * in the request could be stored on a different server.  memcache
   * servers need the ability to query other servers on the users
   * behalf, IMHO. */
  ms = mc_find_server(mc, req->query.tqh_first->hash);
  if (ms == NULL) {
    warnx("%s:%u\tUnable to find a valid server", __FILE__, __LINE__);
    mcFree(rv);
    return;
  }

  if (mc_server_connect(mc, ms) == -1) {
    mcFree(rv);
    return;
  }

  if (writev(ms->fd, rv, num_vec) < 0) {
    warn("%s:%u\twritev()", __FILE__, __LINE__);
    mc_deactivate_server(mc, ms);
    /* XXX Should we recursively attempt to try this query on the
     * remaining servers in the cluster if the writev() fails?
     * Eventually we'd fail once all servers were exhausted?  For now,
     * just fail and return NULL. */
    mcFree(rv);
    return;
  }

  /* Switch to non-blocking io */
  mc_server_block(ms, 1);

  /* Reset our buffer */
  mc_reset_buf(mc);

  /* Grab a line from the server */
  cur = mc_get_line(mc, ms);

  /* Set res to NULL that way we reset our scan for keys from the
   * start of the list. */
  res = NULL;

  /* Loop through reading the server's response until we're done. */
  while(1) {
    if (memcmp(cur, "VALUE ", sizeof("VALUE ") - 1) == 0) {
      /* Found something */
      cur = &cur[sizeof("VALUE ") - 1]; /* advance cursor */

      /* We can do something cool here.  Assume the server is going to
       * return the responses in the same order we requested them.
       * With that presumption being true, start searching from
       * res->tqe_next instead of starting over again. */
      if (res != NULL && res->entries.tqe_next != NULL) {
	for (res = res->entries.tqe_next; res != NULL; res = res->entries.tqe_next) {
	  if (memcmp(cur, res->key, res->len) == 0)
	    break;
	}
      }

      /* If we're still NULL, sequentially scan through the list. */
      if (res == NULL) {
	for (res = req->query.tqh_first; res != NULL; res = res->entries.tqe_next) {
	  if (memcmp(cur, res->key, res->len) == 0)
	    break;
	}

	/* If we're *still* null, then we've got a problem. */
	if (res == NULL) {
	  warnx("%s:%u\tUnable to find a response object for the key %.*s", __FILE__, __LINE__, MAX_KEY_LEN, cur);
	  mcFree(rv);
	  abort();
	}
      }

      cur = &cur[res->len]; /* advance cursor */

      /* Look for the flags */
      flags = (u_int16_t)strtol(cur, &cp, 10);
      if (flags == 0 && (errno == EINVAL || errno == ERANGE)) {
	warn("%s:%u\tstrtol(): invalid flags \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);
	mcFree(rv);
	return;
      }
      res->flags = flags;
      cur = cp; /* advance cursor */

      /* Look for the bytes */
      bytes = (size_t)strtol(cur, &cp, 10);
      if (bytes == 0 && (errno == EINVAL || errno == ERANGE)) {
	warn("%s:%u\tstrtol(): invalid bytes \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);
	mcFree(rv);
	return;
      }
      res->bytes = bytes;
      res->_flags |= MC_RES_FOUND; /* Mark that there was a key found
				    * in memcache, even though it had
				    * * an empty value. */

      /* advance cursor to the start of the value */
      cur = mc->start;

      if (res->size == 0) {
	/* Allocate the necessary response.
	 *
	 * XXX Should think about having an option to mmap(2) an
	 * anonymous region to speed up the read(2) instead of using
	 * mcMalloc(3). */
	res->val = mcMalloc(res->bytes + 1);
	if (res->val == NULL) {
	  mcFree(rv);
	  return;
	}

	res->size = res->bytes + 1;

	/* Throw in a null character at the end of the value for
	 * convenience's sake. */
	cp = (char *)res->val;
	cp[res->bytes] = '\0';
      }

      if (res->bytes + 2 <= mc->size - (size_t)(cur - mc->buf)) {

	/* We were able to read the entire value out of our read(2)
	 * buffer. Cool. */
	memcpy(res->val, mc->start, res->bytes);

	/* Advance the cursor in the event that we were able to fit a
	 * bunch of lines in a single read(2) call. */
	mc->cur = mc->start = &cur[res->bytes + 2];

	/* ...set the FD to be non-blocking again... */
	mc_server_block(ms, 1);

	/* ...and grab a new line so we can start the fun all over
	 * again! */
	cur = mc_get_line(mc, ms);

	continue;
      }

      /* Copy over the remaining bytes from the start of the value to
       * the end of the buffer.  If the value is small enough, we
       * shouldn't need to make a read(2) call to fetch the rest of
       * the data.  Cool. */
      memcpy(res->val, mc->start, res->bytes - mc->size - (size_t)(mc->start - mc->buf));

      /* Before we make another read(2) call, disable non-blocking IO
       * that way the readv(2) call doesn't return until we get the
       * data the server is alegedly sending. */
      mc_server_block(ms, 0);

      /* Since the incoming value is large enough, we need to call
       * read(2) again. */
      cp = (char *)res->val;
      rv[0].iov_base = &cp[res->bytes - mc->size - (size_t)(mc->start - mc->buf)];
      rv[0].iov_len = res->bytes - mc->size - (size_t)(mc->start - mc->buf);

      /* Recycle the buffer at this point. */
#ifdef PEDANTIC
      /* It's safe to assume that our buffer is larger than "\r\n",
       * but it never hurts to check. */
      if (mc->size < sizeof("\r\n") - 1)
	errx(EX_SOFTWARE, "%s:%u\tProtocol Error", __FILE__, __LINE__);
#endif

      rv[1].iov_base = mc->buf;
      rv[1].iov_len = str_endl_len;
      rb = readv(ms->fd, rv, 2);
      switch(rb) {
      case -1:
	warn("%s:%u\tread()", __FILE__, __LINE__);
	mc_deactivate_server(mc, ms);
	mcFree(rv);
	return; /* Should we try and read again depending on errno? */
      case 0:
	/* Seems like an error to me if the server closes its
	 * connection here.  deactivate the server instead of just
	 * closing it. */
	mc_deactivate_server(mc, ms);
	mcFree(rv);
	return;
      }

#ifdef PEDANTIC
      /* Make sure we are where we think we are. */
      if (mc->start[rb - 2] != '\r' && mc->start[rb - 1] != '\n') {
	warnx("%s:%u\tProtocol Error", __FILE__, __LINE__);
	if ((res->_flags & (MC_RES_FREE_ON_DELETE | MC_RES_NO_FREE_ON_DELETE) ==
	     (MC_RES_FREE_ON_DELETE | MC_RES_NO_FREE_ON_DELETE)) ||
	    (res->_flags & MC_RES_FREE_ON_DELETE)) {
	  mcFree(res->val);
	  res->val = NULL;
	  res->size = 0;
	  res->bytes = 0;
	}
	mcFree(rv);
	return;
      }
#endif

      /* Start searching/reading through the buffer from the start
       * again. Since we exceeded the buffer in our prior readv(2)
       * call, reset it... */
      mc_reset_buf(mc);

      /* ...set the FD to be non-blocking again... */
      mc_server_block(ms, 1);

      /* ...and grab a new line so we can start the fun all over
       * again! */
      cur = mc_get_line(mc, ms);
    } else if (memcmp(cur, "END", sizeof("END") - 1) == 0) {
      /* We're done looking for values, or the server didn't find
       * anything. */

      /* ...set the FD to be non-blocking again... */
      mc_server_block(ms, 1);

      mcFree(rv);
      return;
    } else if (memcmp(cur, "ERROR", sizeof("ERROR") - 1) == 0) {
      /* drat, an error: something bad happened */
    } else if (memcmp(cur, "CLIENT_ERROR", sizeof("CLIENT_ERROR") - 1) == 0) {
      err(EX_PROTOCOL, "%s:%u\tClient error", __FILE__, __LINE__);
    } else if (memcmp(cur, "SERVER_ERROR", sizeof("SERVER_ERROR") - 1) == 0) {
      err(EX_PROTOCOL, "%s:%u\tServer error", __FILE__, __LINE__);
    } else {
      err(EX_PROTOCOL, "%s:%u\tUnable to handle response %.*s", __FILE__, __LINE__, 12, mc->buf);
    }
  }
}


void
mc_get(struct memcache *mc, struct memcache_req *req) {
  mc_fetch_cmd(mc, req, str_get_cmd, str_get_len);
}


static char *
mc_get_line(struct memcache *mc, struct memcache_server *ms) {
  ssize_t rb;
  char *cp;
  char *new_start;
#ifdef HAVE_SELECT
  int ret;
#endif

  if (mc->read_cur == NULL) {
    mc->read_cur = mc->buf;
    goto try_read;
  }

  for(;;) {
    /* We're guaranteed to have data in our buffer at this point.
     * Search for a newline.  If we don't have one, we need to either
     * move data around and read(2) again, or mcRealloc(3) and read(2)
     * again. */
    cp = memchr(mc->start, (int)'\n', mc->size - (size_t)(mc->cur - mc->buf));
    if (cp == NULL) {
      get_more_bits:
      /* If our cursor is at the start of the buffer, we need to
       * allocate more memory. */
      if (mc->buf == mc->cur) {
	cp = (char *)mcRealloc(mc->buf, mc->size * 2);
	if (cp == NULL) {
	  /* No sense in continuing if we can't read(2) in a whole
	   * line. There are likely bigger problems on a system if its
	   * experiencing a problem like this. */
	  warn("%s:%u\tmcRealloc()", __FILE__, __LINE__);
	  return NULL;
	}

	mc->buf = mc->cur = mc->start = cp;
	mc->read_cur = &mc->cur[mc->size];
	mc->size *= 2; /* Note the change in buffer size */
      } else {
	/* bcopy(3) the data from mc->cur to the end of mc->buf back
	 * to the start of mc->buf.  Use bcopy(3) in case there is
	 * overlap. */
	bcopy(mc->cur, mc->buf, mc->size - (size_t)(mc->cur - mc->buf));

	/* Advance where the read(2) should place new data. */
	mc->read_cur = &mc->buf[mc->size - (size_t)(mc->cur - mc->buf)];

	/* Reset the cursor to the front of the buffer */
	mc->cur = mc->buf;
      }

      /* Now that we've got space to read(2) in bits, read(2) in
       * data. */
      try_read:

#ifdef HAVE_SELECT
      /* Before we read(2) anything, check to make sure there is data
       * available to be read(2).  No sense in wastefully calling
       * read(2) constantly in a loop. */
      ret = select(1, &ms->fds, NULL, NULL, &ms->select_tv);
      if (ret == -1)
	errx(EX_OSERR, "%s:%u\tselect()", __FILE__, __LINE__);
#endif
      rb = read(ms->fd, mc->read_cur, mc->size - (size_t)(mc->cur - mc->buf));
      switch(rb) {
      case -1:
	/* We're in non-blocking mode, don't abort because of EAGAIN or
	 * EINTR */
	if (errno == EAGAIN || errno == EINTR)
	  goto try_read;
	warn("%s:%u\tread()", __FILE__, __LINE__);
	mc_deactivate_server(mc, ms);
	return NULL; /* Should we try again depending on errno? */
      case 0:
	/* Seems like an error to me if the server closes its connection
	 * here.  deactivate the server instead of just closing it. */
	mc_deactivate_server(mc, ms);
	return NULL;
      }

      /* If we got the same amount of data as we said was max, then we
       * probably need to read(2) more data.  Worst case scenario
       * being that we double our buffer and read(2) returns zero new
       * bytes.  If people are hitting this case a lot,
       * GET_INIT_BUF_SIZE should be increased. */
      if (rb == mc->size - (size_t)(mc->cur - mc->buf)) {
	goto get_more_bits;
      }

      /* Search for a newline again, starting from where we read(2)
       * data in.  Slight optimization over just starting the loop
       * over again since in most cases, we'll find a newline in the
       * newly read(2) chunk of data. */
      cp = memchr(mc->cur, (int)'\n', mc->size - (size_t)(mc->cur - mc->buf));
      if (cp == NULL) {
	/* We suck.  Try again. This is nearly the same as a continue,
	 * but saves us from doing an extra memchr(). */
	goto get_more_bits;
      }
    }

#ifdef PEDANTIC
    /* At this point, we're guaranteed to have a complete line in the
     * buffer.  cp should be pointing to a newline character. */
    if (*cp != '\n') abort();

    /* Protocol check, make sure there's a carige return before the
     * newline */
    if (*(cp - 1) != '\r') {
      warnx("%s:%u\tProtocol violation, no \\r before the \\n", __FILE__, __LINE__);
      mc_deactivate_server(mc, ms);
      return NULL;
    }
#endif

    new_start = cp + 1; /* Advance the start of the next line */
    *(cp - 1) = '\0'; /* Add a null character */
    cp = mc->start;
    mc->start = new_start;

    /* Handy little debugging function: */
    /* warnx("Line reads: \"%.*s\"", (size_t)(mc->start - cp - 2), cp); */
    return cp;
  }

  errx(EX_PROTOCOL, "%s:%u\tProtocol violation: Unable to obtain a new line", __FILE__, __LINE__);
}


#ifdef USE_CRC32_HASH
static const u_int32_t crc32tab[256] = {
  0x00000000, 0x77073096, 0xee0e612c, 0x990951ba,
  0x076dc419, 0x706af48f, 0xe963a535, 0x9e6495a3,
  0x0edb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988,
  0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91,
  0x1db71064, 0x6ab020f2, 0xf3b97148, 0x84be41de,
  0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7,
  0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec,
  0x14015c4f, 0x63066cd9, 0xfa0f3d63, 0x8d080df5,
  0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172,
  0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,
  0x35b5a8fa, 0x42b2986c, 0xdbbbc9d6, 0xacbcf940,
  0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59,
  0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116,
  0x21b4f4b5, 0x56b3c423, 0xcfba9599, 0xb8bda50f,
  0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924,
  0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d,
  0x76dc4190, 0x01db7106, 0x98d220bc, 0xefd5102a,
  0x71b18589, 0x06b6b51f, 0x9fbfe4a5, 0xe8b8d433,
  0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818,
  0x7f6a0dbb, 0x086d3d2d, 0x91646c97, 0xe6635c01,
  0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e,
  0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457,
  0x65b0d9c6, 0x12b7e950, 0x8bbeb8ea, 0xfcb9887c,
  0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65,
  0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2,
  0x4adfa541, 0x3dd895d7, 0xa4d1c46d, 0xd3d6f4fb,
  0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0,
  0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9,
  0x5005713c, 0x270241aa, 0xbe0b1010, 0xc90c2086,
  0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
  0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4,
  0x59b33d17, 0x2eb40d81, 0xb7bd5c3b, 0xc0ba6cad,
  0xedb88320, 0x9abfb3b6, 0x03b6e20c, 0x74b1d29a,
  0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683,
  0xe3630b12, 0x94643b84, 0x0d6d6a3e, 0x7a6a5aa8,
  0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1,
  0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe,
  0xf762575d, 0x806567cb, 0x196c3671, 0x6e6b06e7,
  0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc,
  0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5,
  0xd6d6a3e8, 0xa1d1937e, 0x38d8c2c4, 0x4fdff252,
  0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b,
  0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60,
  0xdf60efc3, 0xa867df55, 0x316e8eef, 0x4669be79,
  0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236,
  0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f,
  0xc5ba3bbe, 0xb2bd0b28, 0x2bb45a92, 0x5cb36a04,
  0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d,
  0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a,
  0x9c0906a9, 0xeb0e363f, 0x72076785, 0x05005713,
  0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38,
  0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21,
  0x86d3d2d4, 0xf1d4e242, 0x68ddb3f8, 0x1fda836e,
  0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777,
  0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c,
  0x8f659eff, 0xf862ae69, 0x616bffd3, 0x166ccf45,
  0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2,
  0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db,
  0xaed16a4a, 0xd9d65adc, 0x40df0b66, 0x37d83bf0,
  0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
  0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6,
  0xbad03605, 0xcdd70693, 0x54de5729, 0x23d967bf,
  0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94,
  0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d,
};
#endif /* USE_CRC32_HASH */

u_int32_t
mc_hash_key(const char *key, const size_t len) {
#ifdef USE_CRC32_HASH
  size_t i;
  u_int32_t crc;

  crc = ~0;

  for (i = 0; i < len; i++)
    crc = (crc >> 8) ^ crc32tab[(crc ^ (key[i])) & 0xff];

  return((~crc >> 16) & 0x7fff);
#else
# ifdef USE_PERL_HASH
  u_int32_t h, i;
  char *p;

  i = len;	/* Work back throug the key length */
  p = key;	/* Character pointer */
  h = 0;	/* The hash value */

  while (i--) {
    h += *p;
    h += (h << 10);
    h ^= (h >> 6);
  }
  h += (h << 3);
  h ^= (h >> 11);
  h += (h << 15);

  return h;
# else
#  ifdef USE_ELF_HASH
  u_int32_t g, h, i;
  char *p;

  i = len;	/* Work back throug the key length */
  p = key;	/* Character pointer */
  h = 0;	/* The hash value */

  while (i--) {
    h = (h << 4) + *p++;
    if (g = h & 0xF0000000)
      h ^= g >> 24;
    h &= ~g;
  }

  return h;
#  else
#   error "Please choose USE_CRC32_HASH, USE_ELF_HASH, or USE_PERL_HASH as a hasing scheme when compiling memcache"
#  endif
# endif
#endif
}


u_int32_t
mc_incr(struct memcache *mc, const char *key, const size_t key_len, const u_int32_t val) {
  return mc_atomic_cmd(mc, str_incr_cmd, str_incr_len, key, key_len, val);
}


struct memcache *
mc_new(void) {
  struct memcache *mc;

  mc = (struct memcache *)mcMalloc(sizeof(struct memcache));
  if (mc != NULL) {
    bzero(mc, sizeof(struct memcache));

    mc->buf = (char *)mcMalloc(GET_INIT_BUF_SIZE);
    if (mc->buf == NULL) {
      mcFree(mc);
      return NULL;
    }
    mc->size = GET_INIT_BUF_SIZE;

    TAILQ_INIT(&mc->server_list);

    /* Set any default values */
    mc->tv.tv_sec = 2;
    mc->tv.tv_usec = 600;
    mc->read_size = 2048;
  }

  return mc;
}


void
mc_refresh(struct memcache *mc, struct memcache_req *req) {
  mc_fetch_cmd(mc, req, str_refresh_cmd, str_refresh_len);
}


int
mc_replace(struct memcache *mc,
	   const char *key, const size_t key_len,
	   const void *val, const size_t bytes,
	   const time_t expire, const u_int16_t flags) {
  return mc_storage_cmd(mc, str_replace_cmd, str_replace_len, key, key_len, val, bytes, expire, flags);
}


struct memcache_res *
mc_req_add(struct memcache_req *req, const char *key, const size_t len) {
  struct memcache_res *res;
  res = mc_res_new();

  res->key = key; /* XXX Should I be mcStrdup()'ing the key? */
  res->len = len;
  res->hash = mc_hash_key(key, len);

  TAILQ_INSERT_TAIL(&req->query, res, entries);
  req->num_keys++;

  return res;
}


void
mc_req_free(struct memcache_req *req) {
  while (req->query.tqh_first != NULL)
    mc_res_free(req, req->query.tqh_first);

  mcFree(req);
}


struct memcache_req *
mc_req_new(void) {
  struct memcache_req *req;

  req = (struct memcache_req *)mcMalloc(sizeof(struct memcache_req));
  if (req != NULL) {
    bzero(req, sizeof(struct memcache_req));

    TAILQ_INIT(&req->query);
  }

  return req;
}


void
mc_res_free(struct memcache_req *req, struct memcache_res *res) {
  TAILQ_REMOVE(&req->query, res, entries);
  if ((res->_flags & (MC_RES_FREE_ON_DELETE | MC_RES_NO_FREE_ON_DELETE) ==
       (MC_RES_FREE_ON_DELETE | MC_RES_NO_FREE_ON_DELETE)) ||
      res->_flags & MC_RES_FREE_ON_DELETE) {
    if (res->size > 0)
      mcFree(res->val);
  }

  mcFree(res);
}


void
mc_res_free_on_delete(struct memcache_res *res, const int fod) {
  if (fod) {
    res->_flags &= ~MC_RES_NO_FREE_ON_DELETE;
    res->_flags |= MC_RES_FREE_ON_DELETE;
  } else {
    res->_flags &= ~MC_RES_FREE_ON_DELETE;
    res->_flags |= MC_RES_NO_FREE_ON_DELETE;
  }
}


static struct memcache_res *
mc_res_new(void) {
  struct memcache_res *res;

  res = (struct memcache_res *)mcMalloc(sizeof(struct memcache_res));
  if (res != NULL) {
    bzero(res, sizeof(struct memcache_res));

    /* Default values */
    res->_flags = MC_RES_FREE_ON_DELETE | MC_RES_NO_FREE_ON_DELETE; /* unset */
  }

  return res;
}


int
mc_server_add(struct memcache *mc, const char *hostname, const char *port) {
  return mc_server_add2(mc, hostname, strlen(hostname), port, strlen(port));
}


int
mc_server_add2(struct memcache *mc, const char *hostname, const size_t hostname_len,
	       const char *port, const size_t port_len) {
  struct memcache_server *ms;
  int ret;

  ms = mc_server_new(mc);
  if (ms == NULL)
    return -1;

  if (hostname == NULL || hostname_len == 0) {
    ms->hostname = mcStrdup("localhost");
  } else {
    ms->hostname = mcMalloc(hostname_len + 1);
    memcpy(ms->hostname, hostname, hostname_len);
    ms->hostname[hostname_len] = '\0';
  }

  if (ms->hostname == NULL) {
    mc_server_free(ms);
    return -2;
  }


  if (port == NULL || port_len == 0) {
    ms->port = mcStrdup("11211");
  } else {
    ms->port = mcMalloc(port_len + 1);
    memcpy(ms->port, port, port_len);
    ms->port[port_len] = '\0';
  }

  if (ms->port == NULL) {
    mc_server_free(ms);
    return -3;
  }

  ret = mc_server_resolve(ms);
  if (ret != 0) {
    warn("libmemcache: host %s does not exist: %s.  Not adding to server list.", ms->hostname, gai_strerror(ret));
    mc_server_free(ms);
    return -4;
  }

  TAILQ_INSERT_TAIL(&mc->server_list, ms, entries);

  /* Add ms to the array of servers to try */
  if (mc->live_servers == NULL) {
    mc->num_live_servers = 1;
    mc->live_servers = (struct memcache_server**)mcMalloc(sizeof(struct memcache_server*) * mc->num_live_servers);
    mc->live_servers[0] = ms;
  } else {
    /* Reallocate mc->live_servers to fit the number of struct
     * memcache_servers entries. */
    mc->num_live_servers++;
    mc->live_servers = (struct memcache_server**)mcRealloc(mc->live_servers, sizeof(struct memcache_server*) * mc->num_live_servers);
    if (mc->live_servers == NULL) {
      warn("libmemcache: Unable to mcRealloc() enough memory to add a new server");
      mc_server_free(ms);
      return -5;
    }

    /* Add the new server to the end of the list */
    mc->live_servers[mc->num_live_servers - 1] = ms;
  }

  return 0;
}


static int
mc_server_connect(struct memcache *mc, struct memcache_server *ms) {
  struct addrinfo *res;
  int i;
#ifdef TCP_NODELAY
  int val;
#endif

  if (ms->fd != -1)
    return ms->fd;

  if (ms->active == 'd' || ms->active == 'n')
    return -1;

  if (ms->hostinfo == NULL || ms->hostinfo->ai_addrlen == 0) {
    i = mc_server_resolve(ms);
    if (i != 0) {
      warn("host %s does not exist: %s.  Not adding to server list.", ms->hostname, gai_strerror(i));
      ms->active = 'n';
      return -1;
    }
  }

  for (i = 0, res = ms->hostinfo; res != NULL; res = res->ai_next, i++) {
    ms->fd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if (ms->fd < 0) {
      warn("%s:%u\tsocket()", __FILE__, __LINE__);
      continue;
    }

#ifdef TCP_NODELAY
    val = 1;
    if (setsockopt(ms->fd, IPPROTO_TCP, TCP_NODELAY, &val, (socklen_t)sizeof(val)) != 0) {
      warn("%s:%u\tsetsockopt(TCP_NODELAY)", __FILE__, __LINE__);
    }
#endif

    if (setsockopt(ms->fd, SOL_SOCKET, SO_SNDTIMEO, &ms->tv, (socklen_t)sizeof(struct timeval)) != 0) {
      warn("%s:%u\tsetsockopt(SO_SNDTIMEO)", __FILE__, __LINE__);
      /* Close the socket, set the file descriptor to -1, and continue
       * trying to connect to the rest of the servers that match this
       * hostname.  More than likely there is only one IP per host
       * name, but, in the event there isn't, continue to the next
       * entry. */
      if (close(ms->fd) != 0)
	warn("%s:%u\tclose()", __FILE__, __LINE__);
      ms->fd = -1;
      continue;
    }

    if (connect(ms->fd, res->ai_addr, (socklen_t)res->ai_addrlen) != 0) {
      warn("%s:%u\tconnect()", __FILE__, __LINE__);
      if (close(ms->fd) != 0)
	warn("%s:%u\tclose()", __FILE__, __LINE__);
      ms->fd = -1;
      continue;
    } else {
#ifdef HAVE_SELECT
      /* Before we return, add our file descriptor to ms->fds. */
      FD_SET(ms->fd, &ms->fds);
#endif
      return ms->fd;
    }
  }

#ifdef PEDANTIC
  if (ms->fd != -1) abort();
#endif

  /* If none of the IP addresses for this hostname work, remove the
   * server from the live_server list (we assume they're live by
   * default) and return -1. */
  mc_deactivate_server(mc, ms);
  return -1;
}


static void
mc_server_free(struct memcache_server *ms) {
  if (ms == NULL)
    return;

  if (ms->hostinfo != NULL)
    freeaddrinfo(ms->hostinfo);

  if (ms->hostname != NULL)
    mcFree(ms->hostname);

  if (ms->port != NULL)
    mcFree(ms->port);

  if (ms->fd != -1) {
    if (close(ms->fd) != 0)
      warn("%s:%u\tclose()", __FILE__, __LINE__);
    ms->fd = -1;
  }

  mcFree(ms);
}


static struct memcache_server *
mc_server_new(struct memcache *mc) {
  struct memcache_server *ms;

  ms = (struct memcache_server *)mcMalloc(sizeof(struct memcache_server));
  if (ms != NULL) {
    bzero(ms, sizeof(struct memcache_server));

    /* Set any default values */
    ms->active = 't';
    ms->fd = -1;
    ms->flags = -1;

    /* Defaults from mc */
    ms->tv.tv_sec = mc->tv.tv_sec;
    ms->tv.tv_usec = mc->tv.tv_usec;
  }

  return ms;
}


static int
mc_server_resolve(struct memcache_server *ms) {
  struct addrinfo hints, *res;
  int ret;

  /* Resolve the hostname ahead of time */
  bzero(&hints, sizeof(struct addrinfo));
  hints.ai_family = PF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  ret = getaddrinfo(ms->hostname, ms->port, &hints, &(ms->hostinfo));
  if (ret != 0)
    return ret;

  for (res = ms->hostinfo; res != NULL; res = res->ai_next)
    ms->num_addrs++;

  return 0;
}


void
mc_server_stats_free(struct memcache_server_stats *s) {
  if (s->version != NULL)
    mcFree(s->version);
  mcFree(s);
}


struct memcache_server_stats *
mc_stats(struct memcache *mc) {
  struct memcache_server *ms;
  struct memcache_server_stats *s, *ts;

  s = mc_server_stats_new();
  for (ms = mc->server_list.tqh_first; ms != NULL; ms = ms->entries.tqe_next) {
    ts = mc_server_stats(mc, ms);
    if (ts == NULL)
      continue;

    /* Merge the values from ts into s.  Any per-server specific data
     * is pulled from the last server. */
    s->pid = ts->pid;
    s->uptime = ts->uptime;
    s->time = ts->time;
    if (s->version == NULL && ts->version != NULL)
      s->version = mcStrdup(ts->version);

    s->rusage_user.tv_sec += ts->rusage_user.tv_sec;
    s->rusage_user.tv_usec += ts->rusage_user.tv_usec;
    if (s->rusage_user.tv_usec > 1000000) {
      s->rusage_user.tv_sec += s->rusage_user.tv_usec / 1000000;
      s->rusage_user.tv_usec -= 1000000 * (s->rusage_user.tv_usec / 1000000);
    }

    s->rusage_system.tv_sec += ts->rusage_system.tv_sec;
    s->rusage_system.tv_usec += ts->rusage_system.tv_usec;
    if (s->rusage_system.tv_usec > 1000000) {
      s->rusage_system.tv_sec += s->rusage_system.tv_usec / 1000000;
      s->rusage_system.tv_usec -= 1000000 * (s->rusage_system.tv_usec / 1000000);
    }

    s->curr_items += ts->curr_items;
    s->total_items += ts->total_items;
    s->bytes = s->bytes + ts->bytes;
    s->curr_connections += ts->curr_connections;
    s->total_connections += ts->total_connections;
    s->connection_structures += ts->connection_structures;
    s->cmd_get += ts->cmd_get;
    s->cmd_set += ts->cmd_set;
    s->get_hits += ts->get_hits;
    s->get_misses += ts->get_misses;
    s->bytes_read += ts->bytes_read;
    s->bytes_written += ts->bytes_written;
    s->limit_maxbytes += ts->limit_maxbytes;

    mc_server_stats_free(ts);
  }

  return s;
}


static struct memcache_server_stats *
mc_server_stats_new(void) {
  struct memcache_server_stats *s;
  s = (struct memcache_server_stats *)mcMalloc(sizeof(struct memcache_server_stats));
  if (s != NULL) {
    bzero(s, sizeof(struct memcache_server_stats));
  }

  return s;
}


struct memcache_server_stats *
mc_server_stats(struct memcache *mc, struct memcache_server *ms) {
  struct memcache_server_stats *s;
  char *cp, *cur;

  if (mc_server_connect(mc, ms) == -1)
    return NULL;

  if (write(ms->fd, "stats\r\n", sizeof("stats\r\n") - 1) < 0) {
    warn("%s:%u\twrite()", __FILE__, __LINE__);
    mc_deactivate_server(mc, ms);
    return NULL;
  }

  s = mc_server_stats_new();
  if (s == NULL)
    return NULL;

  /* Switch to non-blocking io */
  mc_server_block(ms, 1);

  /* Reset our buffer */
  mc_reset_buf(mc);

  for(;;) {
    cur = mc_get_line(mc, ms);

    if (cur != NULL && memcmp(cur, "STAT ", sizeof("STAT ") - 1) == 0) {
      cur = &cur[sizeof("STAT ") - 1];

      /* Time to loop through the potential stats keys.  Joy.  This is
       * going to complete in O(1 + 2 + 3 ... N) operations (currently
       * 190).  Ugh.  Don't know of a better way to handle this
       * without a hash.  Besides, this is just stats. */
      if (memcmp(cur, "pid ", sizeof("pid ") - 1) == 0) {
	cur = &cur[sizeof("pid ") - 1];
	s->pid = (pid_t)strtol(cur, &cp, 10);
	if (s->pid == 0 && (errno == EINVAL || errno == ERANGE))
	  err(EX_PROTOCOL, "%s:%u\tstrtol(): invalid pid \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);
      } else if (memcmp(cur, "uptime ", sizeof("uptime ") - 1) == 0) {
	cur = &cur[sizeof("uptime ") - 1];
	s->uptime = (time_t)strtol(cur, &cp, 10);
	if (s->uptime == 0 && (errno == EINVAL || errno == ERANGE))
	  err(EX_PROTOCOL, "%s:%u\tstrtol(): invalid uptime \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);
      } else if (memcmp(cur, "time ", sizeof("time ") - 1) == 0) {
	cur = &cur[sizeof("time ") - 1];
	s->time = (time_t)strtol(cur, &cp, 10);
	if (s->time == 0 && (errno == EINVAL || errno == ERANGE))
	  err(EX_PROTOCOL, "%s:%u\tstrtol(): invalid time \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);
      } else if (memcmp(cur, "version ", sizeof("version ") - 1) == 0) {
	cur = &cur[sizeof("version ") - 1];
	for (cp = cur; !isspace(*cp); cp++);
	s->version = (char *)mcMalloc((size_t)(cp - cur + 1));
	if (s->version == NULL) {
	  warn("%s:%u\tmcMalloc()", __FILE__, __LINE__);
	} else {
	  memcpy(s->version, cur, (size_t)(cp - cur));
	  s->version[(size_t)(cp - cur)] = '\0';
	}
      } else if (memcmp(cur, "rusage_user ", sizeof("rusage_user ") - 1) == 0) {
	cur = &cur[sizeof("rusage_user ") - 1];
	s->rusage_user.tv_sec = (int32_t)strtol(cur, &cp, 10);
	if (s->rusage_user.tv_sec == 0 && (errno == EINVAL || errno == ERANGE))
	  err(EX_PROTOCOL, "%s:%u\tstrtol(): invalid rusage_user seconds \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);
	else {
	  cur = cp; /* advance cursor */
#ifdef PEDANTIC
	  if (!(*cur == '.' || *cur == ':'))
	    warn("%s:%u\tProtocol violation: invalid separator: %x", __FILE__, __LINE__, *cur);
	  else {
#endif
	    cur++; /* advance past colon */
	    s->rusage_user.tv_usec = (int32_t)strtol(cur, &cp, 10);
	    if (s->rusage_user.tv_usec == 0 && (errno == EINVAL || errno == ERANGE))
	      err(EX_PROTOCOL, "%s:%u\tstrtol(): invalid rusage_user microseconds \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);
#ifdef PEDANTIC
	  }
#endif
	}

      } else if (memcmp(cur, "rusage_system ", sizeof("rusage_system ") - 1) == 0) {
	cur = &cur[sizeof("rusage_system ") - 1];
	s->rusage_system.tv_sec = (int32_t)strtol(cur, &cp, 10);
	if (s->rusage_system.tv_sec == 0 && (errno == EINVAL || errno == ERANGE))
	  err(EX_PROTOCOL, "%s:%u\tstrtol(): invalid rusage_system seconds \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);
	else {
	  cur = cp; /* advance cursor */
#ifdef PEDANTIC
	  if (!(*cur == '.' || *cur == ':'))
	    err(EX_PROTOCOL, "%s:%u\tProtocol violation: invalid separator: %x", __FILE__, __LINE__, *cur);
	  else {
#endif
	    cur++; /* advance past colon */
	    s->rusage_system.tv_usec = (int32_t)strtol(cur, &cp, 10);
	    if (s->rusage_system.tv_usec == 0 && (errno == EINVAL || errno == ERANGE))
	      err(EX_PROTOCOL, "%s:%u\tstrtol(): invalid rusage_system microseconds \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);
#ifdef PEDANTIC
	  }
#endif
	}
      } else if (memcmp(cur, "curr_items ", sizeof("curr_items ") - 1) == 0) {
	cur = &cur[sizeof("curr_items ") - 1];
	s->curr_items = (u_int32_t)strtol(cur, &cp, 10);
	if (s->curr_items == 0 && (errno == EINVAL || errno == ERANGE))
	  err(EX_PROTOCOL, "%s:%u\tstrtol(): invalid curr_items \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);
      } else if (memcmp(cur, "total_items ", sizeof("total_items ") - 1) == 0) {
	cur = &cur[sizeof("total_items ") - 1];
	s->total_items = (u_int64_t)strtoll(cur, &cp, 10);
	if (s->total_items == 0 && (errno == EINVAL || errno == ERANGE))
	  err(EX_PROTOCOL, "%s:%u\tstrtoll(): invalid total_items \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);
      } else if (memcmp(cur, "bytes ", sizeof("bytes ") - 1) == 0) {
	cur = &cur[sizeof("bytes") - 1];
	s->bytes = (u_int64_t)strtoll(cur, &cp, 10);
	if (s->bytes == 0 && (errno == EINVAL || errno == ERANGE))
	  err(EX_PROTOCOL, "%s:%u\tstrtol(): invalid bytes \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);
      } else if (memcmp(cur, "curr_connections ", sizeof("curr_connections ") - 1) == 0) {
	cur = &cur[sizeof("curr_connections ") - 1];
	s->curr_connections = (u_int32_t)strtol(cur, &cp, 10);
	if (s->curr_connections == 0 && (errno == EINVAL || errno == ERANGE))
	  err(EX_PROTOCOL, "%s:%u\tstrtol(): invalid curr_connections \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);
      } else if (memcmp(cur, "total_connections ", sizeof("total_connections ") - 1) == 0) {
	cur = &cur[sizeof("total_connections ") - 1];
	s->total_connections = (u_int64_t)strtoll(cur, &cp, 10);
	if (s->total_connections == 0 && (errno == EINVAL || errno == ERANGE))
	  err(EX_PROTOCOL, "%s:%u\tstrtoll(): invalid total_connections \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);
      } else if (memcmp(cur, "connection_structures ", sizeof("connection_structures ") - 1) == 0) {
	cur = &cur[sizeof("connection_structures ") - 1];
	s->connection_structures = (u_int32_t)strtol(cur, &cp, 10);
	if (s->connection_structures == 0 && (errno == EINVAL || errno == ERANGE))
	  err(EX_PROTOCOL, "%s:%u\tstrtol(): invalid connection_structures \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);
      } else if (memcmp(cur, "cmd_get ", sizeof("cmd_get ") - 1) == 0) {
	cur = &cur[sizeof("cmd_get ") - 1];
	s->cmd_get = (u_int64_t)strtoll(cur, &cp, 10);
	if (s->cmd_get == 0 && (errno == EINVAL || errno == ERANGE))
	  err(EX_PROTOCOL, "%s:%u\tstrtoll(): invalid cmd_get \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);
      } else if (memcmp(cur, "cmd_set ", sizeof("cmd_set ") - 1) == 0) {
	cur = &cur[sizeof("cmd_set ") - 1];
	s->cmd_set = (u_int64_t)strtoll(cur, &cp, 10);
	if (s->cmd_set == 0 && (errno == EINVAL || errno == ERANGE))
	  err(EX_PROTOCOL, "%s:%u\tstrtoll(): invalid cmd_set \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);
      } else if (memcmp(cur, "get_hits ", sizeof("get_hits ") - 1) == 0) {
	cur = &cur[sizeof("get_hits ") - 1];
	s->get_hits = (u_int64_t)strtoll(cur, &cp, 10);
	if (s->get_hits == 0 && (errno == EINVAL || errno == ERANGE))
	  err(EX_PROTOCOL, "%s:%u\tstrtoll(): invalid get_hits \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);
      } else if (memcmp(cur, "get_misses ", sizeof("get_misses ") - 1) == 0) {
	cur = &cur[sizeof("get_misses ") - 1];
	s->get_misses = (u_int64_t)strtoll(cur, &cp, 10);
	if (s->get_misses == 0 && (errno == EINVAL || errno == ERANGE))
	  err(EX_PROTOCOL, "%s:%u\tstrtoll(): invalid get_misses \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);
      } else if (memcmp(cur, "bytes_read ", sizeof("bytes_read ") - 1) == 0) {
	cur = &cur[sizeof("bytes_read ") - 1];
	s->bytes_read = (u_int64_t)strtoll(cur, &cp, 10);
	if (s->bytes_read == 0 && (errno == EINVAL || errno == ERANGE))
	  err(EX_PROTOCOL, "%s:%u\tstrtoll(): invalid bytes_read \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);
      } else if (memcmp(cur, "bytes_written ", sizeof("bytes_written ") - 1) == 0) {
	cur = &cur[sizeof("bytes_written ") - 1];
	s->bytes_written = (u_int64_t)strtoll(cur, &cp, 10);
	if (s->bytes_written == 0 && (errno == EINVAL || errno == ERANGE))
	  err(EX_PROTOCOL, "%s:%u\tstrtoll(): invalid bytes_written \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);
      } else if (memcmp(cur, "limit_maxbytes ", sizeof("limit_maxbytes ") - 1) == 0) {
	cur = &cur[sizeof("limit_maxbytes ") - 1];
	s->limit_maxbytes = (u_int64_t)strtoll(cur, &cp, 10);
	if (s->limit_maxbytes == 0 && (errno == EINVAL || errno == ERANGE))
	  err(EX_PROTOCOL, "%s:%u\tstrtoll(): invalid limit_maxbytes \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);
      } else {
	for (cp = cur; !isspace(*cp); cp++);
	warn("%s:%u\tProtocol error: unknown stat \"%.*s\"", __FILE__, __LINE__, (int)(cp - cur), cur);
      }

      /* Now that we've sucked in our stats value, set our cursor to
       * the end of the value. */
      cp = memchr(cur, (int)'\r', mc->size - (size_t)(cur - mc->buf));
      if (cp == NULL || cp[1] != '\n') {
	warnx("Protocol error: anticipated end of stats value: not at end of stats value");
	mc_server_stats_free(s);
	mc_deactivate_server(mc, ms);
	return NULL;
      }
    } else if (cur != NULL && memcmp(cur, "END", sizeof("END") - 1) == 0) {
      /* We're done reading in stats. */
      break;
    } else {
      errx(EX_PROTOCOL, "%s:%u\tUnable to handle response: \"%.*s\"", __FILE__, __LINE__, 15, cur);
    }
  }

  /* Switch to blocking io */
  mc_server_block(ms, 0);

  return s;
}


void
mc_timeout(struct memcache *mc, int sec, int msec) {
  mc->tv.tv_sec = sec;
  mc->tv.tv_usec = msec;
}


static void
mc_reset_buf(struct memcache *mc) {
  /* Reset our cursor to the front of the buffer */
  mc->cur = mc->start = mc->buf;
  mc->read_cur = NULL;
}


static void
mc_server_block(struct memcache_server *ms, int use_nbio) {
  if (ms->flags == -1) {
    ms->flags = fcntl(ms->fd, F_GETFL, 0);
    if (ms->flags == -1)
      err(EX_OSERR, "%s:%u\tfcntl(F_GETFL)", __FILE__, __LINE__);
  }

  if (use_nbio != 0) {
    /* Switch to non-blocking io */
    if ((ms->flags & O_NDELAY) != 0) {
      ms->flags |= O_NDELAY;
      goto set_flags;
    }
  } else {
    if (ms->flags & O_NDELAY) {
      ms->flags &= ~O_NDELAY;
      goto set_flags;
    }
  }
  return;

  set_flags:
  if (fcntl(ms->fd, F_SETFL, ms->flags) < 0)
    err(EX_OSERR, "%s:%u\tfcntl(F_SETFL)", __FILE__, __LINE__);
}


int
mc_set(struct memcache *mc,
       const char *key, const size_t key_len,
       const void *val, const size_t bytes,
       const time_t expire, const u_int16_t flags) {
  return mc_storage_cmd(mc, str_set_cmd, str_set_len, key, key_len, val, bytes, expire, flags);
}


static int
mc_storage_cmd(struct memcache *mc,
	       const char *cmd, const size_t cmd_len,
	       const char *key, const size_t key_len,
	       const void *val, const size_t bytes,
	       const time_t expire, const u_int16_t flags) {
  struct memcache_server *ms;
  u_int32_t hash;
  char *cur;
  size_t buf_left, i;
  struct iovec wv[11];

  hash = mc_hash_key(key, key_len);
  ms = mc_find_server(mc, hash);
  if (ms == NULL) {
    warnx("%s:%u\tUnable to find a valid server", __FILE__, __LINE__);
    return -1;
  }

  if (mc_server_connect(mc, ms) == -1)
    return -2;

  /* Reset the buffer so that I can chop it up to use it as a scratch
   * pad for converting numbers from binary to ascii.  *hears the baby
   * jesus crying* */
  mc_reset_buf(mc);
  cur = mc->buf;
  buf_left = mc->size;

  wv[0].iov_base = cmd;
  wv[0].iov_len = cmd_len;
  wv[1].iov_base = key;
  wv[1].iov_len = key_len;
  wv[2].iov_base = str_space;
  wv[2].iov_len = str_space_len;

  /* Convert flags to a string */
  i = (size_t)snprintf(cur, buf_left, "%hu", flags);
  if (i < 1) {
    warnx("%s:%u\tsnprintf()", __FILE__, __LINE__);
    return -3;
  }

  wv[3].iov_base = cur;		/* Note where our flags string is located */
  wv[3].iov_len = i;
  wv[4].iov_base = wv[2].iov_base;	/* Add a space */
  wv[4].iov_len = wv[2].iov_len;

  cur = &cur[++i];	/* advance cursor past trailing '\0' */
  buf_left -= i;	/* Note our consumption of some buffer */

  /* Convert expiration time to a string */
  i = (size_t)snprintf(cur, buf_left, "%lu", expire);
  if (i < 1) {
    warnx("%s:%u\tsnprintf()", __FILE__, __LINE__);
    return -4;
  }

  wv[5].iov_base = cur;		/* Note where our flags string is located */
  wv[5].iov_len = i;
  wv[6].iov_base = wv[2].iov_base;	/* Add a space */
  wv[6].iov_len = wv[2].iov_len;

  cur = &cur[++i];	/* advance cursor past trailing '\0' */
  buf_left -= i;	/* Note our consumption of some buffer */

  /* Convert bytes to a string */
  i = (size_t)snprintf(cur, buf_left, "%zu", bytes);
  if (i < 1) {
    warnx("%s:%u\tsnprintf()", __FILE__, __LINE__);
    return -5;
  }

  wv[7].iov_base = cur;	/* Note where our flags string is located */
  wv[7].iov_len = i;
  wv[8].iov_base = str_endl;	/* Newline */
  wv[8].iov_len = str_endl_len;

  cur = &cur[++i];	/* advance cursor past trailing '\0' */
  buf_left -= i;	/* Note our consumption of some buffer */

  /* Add the data */
  wv[9].iov_base = val;
  wv[9].iov_len = bytes;

  /* Add another carrige return */
  wv[10].iov_base = str_endl;
  wv[10].iov_len = str_endl_len;

  if (writev(ms->fd, wv, 11) < 0) {
    warn("%s:%u\twritev()", __FILE__, __LINE__);
    mc_deactivate_server(mc, ms);
    /* XXX Should we recursively attempt to try this query on the
     * remaining servers in the cluster if the writev() fails?
     * Eventually we'd fail once all servers were exhausted?  For now,
     * just fail and return NULL. */
    return -6;
  }

  mc_server_block(ms, 1);
  mc_reset_buf(mc);
  cur = mc_get_line(mc, ms);
  if (cur != NULL && memcmp(cur, "STORED", sizeof("STORED") - 1) == 0) {
    /* Groovy Tuesday */
    mc_server_block(ms, 0);
    return 0;
  } else if (cur != NULL && memcmp(cur, "NOT_STORED", sizeof("NOT_STORED") - 1) == 0) {
    /* Fuck beans.  That was them, wasn't it? */
    mc_server_block(ms, 0);
    warnx("%s:%u\tUnable to store", __FILE__, __LINE__);
    return 1;
  }
  warnx("%s:%u\tProtocol error: %s", __FILE__, __LINE__, cur);
  return 2;
}
